var fs = require("fs")
var lob = require("lob-enc")
var crc = require("crc")


// Expanded bitmasks of nRF51 Image Types
const nrf51ImageTypeNone = 0;
const nrf51ImageTypeSoftdevice = 1;
const nrf51ImageTypeBootloader = 2;
const nrf51ImageTypeSoftdeviceAndBootloader = 3;
const nrf51ImageTypeApplication = 4;
const nrf51ImageTypeBootloaderAndApplication = 6;
const nrf51ImageTypeSoftdeviceAndBootloaderAndApplication = 7;

const STM32_DFU_LOCATION = 0x180000;
const FALLBACK_DFU_LOCATION = 0xDC000;
const BOOTLOADER_DFU_LOCATION = 0x0DA000;
const L21_DFU_LOCATION = 0x1C0000;

var DFU = module.exports = {
  patchos : function dfu(link, firmware, tick)
  {
    return new Promise((res, rej) => {
    var json = {type:'dfu'};
    var channel = link.channel({json:json});
    var start = Date.now();
    var offset = 0;
    var size = 1024;
    var checksum;
    var totalLen = 0;
    var done = false;
    channel.receiving = function(err, packet, cbChan)
    {
      //console.log("We got a packet on the dfu channel\n");
      // any responce is next frame
      if(packet && packet.json && packet.json.end)
      {
        //console.log("got end")
        res();

        return;
      }

      if(packet && !done)
      {
        tick()
        var json = {at:offset};
        var readlength ;
        if (firmware.length <= 1024){
          readlength = firmware.length;
        } else {
          readlength = 1024;
        }

        var body = firmware.slice(0, readlength);
        firmware = firmware.slice(readlength);
  //      //console.log("dfu chunk:", body);
        offset += readlength;
        //buf.copy(body,0,0,body.length);
        checksum = crc.crc32(body,checksum);
        // if at the end, include total checksum
        if(firmware.length === 0)
        {
          ////console.log("done, allowing time for reset: ", totalLen);
          //console.log("dfu done, wait for reset...")
          res()
          json.done = checksum;
          done = true;
          //json.end = true;
          //fs.closeSync(fd);
          //fd = undefined
        }
        //process.stdout.write(".")
        //console.log('\n\ndfu frame',json,readlength, body);
        channel.send({json:json,body:body});
      }else{
          tick();
          res()

      }
      cbChan();
    }
    //console.log('\n\ndfu initiating',link.hashname);
    channel.send({json:json});
    })

  },


  tapos : function dfu_stm32(link,firmware, tick)
  {
    return new Promise((res, rej) => {
    var json = {type:'dfu-stm32'};
    var channel = link.channel({json:json});
    var start = Date.now();
    var offset = 0;
    var size = 1024;
    var checksum;
    var done = false;
    channel.receiving = function(err, packet, cbChan)
    {
  //    //console.log("We got a packet on the dfu-stm32 channel\n");
      // any responce is next frame
      if(packet && !done)
      {
        tick();
        var json = {at:offset};
        var readlength;
        if (firmware.length <= 1024){
          readlength = firmware.length;
        } else {
          readlength = 1024;
        }

        var body = firmware.slice(0, readlength);
        firmware = firmware.slice(readlength);
  //      //console.log("dfu-stm32 chunk:", body);
        offset += readlength;
        //buf.copy(body,0,0,body.length);
        checksum = crc.crc32(body,checksum);
        // if at the end, include total checksum
        if(firmware.length === 0)
        {
          console.log("done")
          done = true;
          json.done = checksum;
          json.end = true;
        }
        //process.stdout.write(".")
  //      //console.log('\n\ndfu-stm32 frame',json,readlength, body);
        channel.send({json:json,body:body});
      }else{
        tick();
        //console.log("dfu-stm32 done, wait for reset...")
          res()
      }
      cbChan();
    }
    //console.log('\n\ndfu-stm32 initiating',link.hashname);
    channel.send({json:json});
    })

  },

  foible_application : function ble(link,firmware, tick)
  {
    return new Promise((res, rej) => {
      var json = {type:'dfu-nrf51'};
    var channel = link.channel({json:json});
    var start = Date.now();
    var imageSizeTotal = firmware.length;
//    console.log("DFU-NRF51 opening ", firmware, " of type ", imageType, ", size ", imageSizeTotal);
    var offset = 0;
    var size = 1024;
    var checksum;
    var done = false;

    var imageType = nrf51ImageTypeApplication;

    // Validate image sizing
    // If we have a single image, just use the file size
    // Otherwise we need to make sure the sizes add up to the file provided
    if (imageType == nrf51ImageTypeNone) {
      //console.log("DFU-NRF51 no valid image type provided");
      return 0;
    } else if (imageType == nrf51ImageTypeSoftdevice) {
      imageSizeSoftdevice = imageSizeTotal;
    } else if (imageType == nrf51ImageTypeBootloader) {
      imageSizeBootloader = imageSizeTotal;
    } else if (imageType == nrf51ImageTypeApplication) {
      imageSizeApplication = imageSizeTotal;
    } else {
      // Multiple images
      // Only use sizes for images specified in type
      imageSizeSoftdevice = (imageType & nrf51ImageTypeSoftdevice) ? imageSizeSoftdevice : 0;
      imageSizeBootloader = (imageType & nrf51ImageTypeBootloader) ? imageSizeBootloader : 0;
      imageSizeApplication = (imageType & nrf51ImageTypeApplication) ? imageSizeApplication : 0;
      if (imageSizeTotal != (imageSizeSoftdevice + imageSizeBootloader + imageSizeApplication)) {
        //console.log("DFU-NRF51 Invalid image sizes provided; does not add up to file size");
        //console.log("TOTAL: ", imageSizeTotal, imageSizeSoftdevice + imageSizeBootloader + imageSizeApplication)
        //console.log("SD: ", imageSizeSoftdevice, " BL: ", imageSizeBootloader, " APP: ", imageSizeApplication);
        return 0;
      }
    }


    channel.receiving = function(err, packet, cbChan)
    {
  //    //console.log("We got a packet on the dfu-nrf51 channel\n");
      // any responce is next frame
      if(packet && !done)
      {
        tick();
        var json = {at:offset};
        var readlength ;
        if (firmware.length <= 1024){
          readlength = firmware.length;
        } else {
          readlength = 1024;
        }

        var body = firmware.slice(0, readlength);
        firmware = firmware.slice(readlength);
        // On the first frame include the image sizes and type
        if (offset == 0) {
          json.start = true;
          if(imageType & nrf51ImageTypeSoftdevice) {
            json.imageSizeSoftdevice = imageSizeSoftdevice;
          }
          if(imageType & nrf51ImageTypeBootloader) {
            json.imageSizeBootloader = imageSizeBootloader;
          }
          if(imageType & nrf51ImageTypeApplication) {
            json.imageSizeApplication = imageSizeApplication;
          }
          json.imageType = imageType;
        }
  //      //console.log("dfu-nrf51 chunk:", body);
        offset += readlength;

        checksum = crc.crc32(body,checksum);
        // if at the end, include total checksum and image type
        if(firmware.length === 0)
        {
          json.done = checksum;
          json.end = true;
          done = true;
        }
        //process.stdout.write(".")
//        console.log('\n\ndfu-nrf51 frame',json,readlength, body.length);
        channel.send({json:json,body:body});
      }else{
        res()
        //console.log('\n\ndfu-nrf51 frame error',err);
      }
      cbChan();
    }
    //console.log('\n\ndfu-nrf51 initiating',link.hashname);
    channel.send({json:json});
    })

  }
}
